#ifndef __mVec3_H__
#define __mVec3_H__

/*! 
    \file mVec3.h
    \brief 3D Vector
    \author Douglas H. Cox
*/

//============================================================================
//    REQUIRED HEADERS
//============================================================================
#include "mMath.h"

//============================================================================
//    DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS / CLASS FORWARDS
//============================================================================
class mFrame;
class mOCS;
//============================================================================
//    FUNCTION PROTOTYPES
//============================================================================

//============================================================================
//    CLASS PROTOTYPES
//============================================================================

//! 3D Vector / Point
/*! mVec3 represents a vector consisting of an array of three float values. */
class MMATH_DLL mVec3
{
public:
  float v[3];   //!< vector data

  //! Default constructor
  /*! No initialization is done here. */
  mVec3() {}

  //! Copy constructor
  mVec3(const mVec3& src)
  { 
    v[0] = src.v[0]; 
    v[1] = src.v[1]; 
    v[2] = src.v[2];
  }

  //! Constructor from 3 float values
  mVec3(float x, float y, float z)
  {
    v[0] = x; 
    v[1] = y; 
    v[2] = z;
  }

  //! Constructor from an array of 3 float values
  mVec3(float *src)
  {
    v[0] = src[0];
    v[1] = src[1];
    v[2] = src[2];
  }

  //! Assignment
  void set(const mVec3& src)
  {
    v[0] = src.v[0]; 
    v[1] = src.v[1]; 
    v[2] = src.v[2];
  }

  //! Assignment from 3 float values
  void set(float x, float y, float z)
  {
    v[0] = x; 
    v[1] = y; 
    v[2] = z;
  }

  //! Assignment from an arry of 3 float values
  void set(float *src)
  {
    v[0] = src[0];
    v[1] = src[1];
    v[2] = src[2];
  }

  //! Assignment
  mVec3& operator = (const mVec3& src)
  { 
    v[0] = src.v[0]; 
    v[1] = src.v[1]; 
    v[2] = src.v[2];
    return *this;
  }

  //! Sets this to a vector pointing in the specified heading.
  /*! The resulting vector is normalized. */
  void setH(float heading)
  {
    mSinCos(heading, v[0], v[1]);
    v[0] = -v[0];
    v[2] = 0.0f;
  }

  //! Sets this to a vector pointing in the specified heading and pitch.
  /*! The resulting vector is normalized. */
  void setHP(float heading, float pitch);

  //! Accessor
  void get(float& x, float& y, float& z) const
  {
    x = v[0];
    y = v[1];
    z = v[2];
  }

  //! Returns the heading and pitch for a normalized vector.
  void getHP(float& heading, float& pitch) const;

  //! Accessor
  float& operator [] (int i)
  {
    return v[i];
  }

  //! Accessor
  float operator [] (int i) const
  {
    return v[i];
  }

  //! Returns true if this == a
  bool operator == (const mVec3& a) const
  {
    return( v[0] == a.v[0] &&
            v[1] == a.v[1] &&
            v[2] == a.v[2] );
  }

  //! Returns true if this != a
  bool operator != (const mVec3& a) const
  { 
    return !(operator==(a));
  }

  //! Returns true is this is within some tolerance value of a.
  bool approxEqual(const mVec3& a, float tol = mZeroTol) const
  {
    return( mApproxEqual(v[0], a.v[0], tol) &&
            mApproxEqual(v[1], a.v[1], tol) &&
            mApproxEqual(v[2], a.v[2], tol) );
  }

  //! Sets each component to the min of this and a.
  void setMin(const mVec3& a)
  {
    if (a.v[0] < v[0]) v[0] = a.v[0];
    if (a.v[1] < v[1]) v[1] = a.v[1];
    if (a.v[2] < v[2]) v[2] = a.v[2];
  }

  //! Sets each component to the max of this and a.
  void setMax(const mVec3& a)
  {
    if (a.v[0] > v[0]) v[0] = a.v[0];
    if (a.v[1] > v[1]) v[1] = a.v[1];
    if (a.v[2] > v[2]) v[2] = a.v[2];
  }

  //! Clamps each component of this to [a, b]
  void clamp(float a, float b)
  {
    v[0] = mClamp(v[0], a, b);
    v[1] = mClamp(v[1], a, b);
    v[2] = mClamp(v[2], a, b);
  }

  //! Clamps each component of vec to [a, b]
  void clamp(const mVec3& vec, float a, float b)
  {
    v[0] = mClamp(vec.v[0], a, b);
    v[1] = mClamp(vec.v[1], a, b);
    v[2] = mClamp(vec.v[2], a, b);
  }

  //! Addition: this += a
  void add(const mVec3& a)
  {
    v[0] += a.v[0];
    v[1] += a.v[1];
    v[2] += a.v[2];
  }

  //! Addition: this = a + b
  void add(const mVec3& a, const mVec3& b)
  {
    v[0] = a.v[0] + b.v[0];
    v[1] = a.v[1] + b.v[1];
    v[2] = a.v[2] + b.v[2];
  }

  //! Addition: Returns this + a
  mVec3 operator + (const mVec3& a) const
  {
    return mVec3(v[0] + a.v[0],
                 v[1] + a.v[1],
                 v[2] + a.v[2]);
  }

  //! Addition: this += a
  mVec3& operator += (const mVec3& a)
  {
    v[0] += a.v[0];
    v[1] += a.v[1];
    v[2] += a.v[2];
    return *this;
  }

  //! this += a * s
  void addScaled(const mVec3& a, float s)
  {
    v[0] += a.v[0] * s;
    v[1] += a.v[1] * s;
    v[2] += a.v[2] * s;
  }

  //! this = a + b * s
  void addScaled(const mVec3& a, const mVec3& b, float s)
  {
    v[0] = a.v[0] + b.v[0] * s;
    v[1] = a.v[1] + b.v[1] * s;
    v[2] = a.v[2] + b.v[2] * s;
  }

  //! this = a * s1 + b * s2
  void addScaled(const mVec3& a, float s1,
                 const mVec3& b, float s2)
  {
    v[0] = a.v[0] * s1 + b.v[0] * s2;
    v[1] = a.v[1] * s1 + b.v[1] * s2;
    v[2] = a.v[2] * s1 + b.v[2] * s2;
  }

  //! Subtraction: this -= a
  void sub(const mVec3& a)
  {
    v[0] -= a.v[0];
    v[1] -= a.v[1];
    v[2] -= a.v[2];
  }

  //! Subtraction: this = a - b
  void sub(const mVec3& a, const mVec3& b)
  {
    v[0] = a.v[0] - b.v[0];
    v[1] = a.v[1] - b.v[1];
    v[2] = a.v[2] - b.v[2];
  }

  //! Subtraction: Returns this - a
  mVec3 operator - (const mVec3& a) const
  {
    return mVec3(v[0] - a.v[0],
                 v[1] - a.v[1],
                 v[2] - a.v[2]);
  }

  //! Subtraction: this -= a
  mVec3& operator -= (const mVec3& a)
  {
    v[0] -= a.v[0];
    v[1] -= a.v[1];
    v[2] -= a.v[2];
    return *this;
  }

  //! Scales each component by s.
  void scale(float s)
  {
    v[0] *= s; 
    v[1] *= s; 
    v[2] *= s;
  }

  //! Scales each component of a by s.
  void scale(const mVec3& a, float s)
  {
    v[0] = a.v[0] * s;
    v[1] = a.v[1] * s;
    v[2] = a.v[2] * s;
  }

  //! Scales each component by s.
  mVec3& operator *= (float s)
  {
    v[0] *= s; 
    v[1] *= s; 
    v[2] *= s;
    return *this;
  }

  //! Negate: this = -this
  void neg()
  {
    v[0] = -v[0];
    v[1] = -v[1];
    v[2] = -v[2];
  }

  //! Negate: this = -a
  void neg(const mVec3& a)
  {
    v[0] = -a.v[0];
    v[1] = -a.v[1];
    v[2] = -a.v[2];
  }

  //! Negate: Returns -this
  mVec3 operator - () const
  {
    return mVec3(-v[0], -v[1], -v[2]);
  }

  //! Component multiplication
  void compMult(const mVec3& a)
  {
    v[0] *= a[0];
    v[1] *= a[1];
    v[2] *= a[2];
  }

  //! Component multiplication
  void compMult(const mVec3& a, const mVec3& b)
  {
    v[0] = a[0] * b[0];
    v[1] = a[1] * b[1];
    v[2] = a[2] * b[2];
  }

  //! Component division
  void compDiv(const mVec3& a)
  {
    v[0] /= a[0];
    v[1] /= a[1];
    v[2] /= a[2];
  }

  //! Component division
  void compDiv(const mVec3& a, const mVec3& b)
  {
    v[0] = a[0] / b[0];
    v[1] = a[1] / b[1];
    v[2] = a[2] / b[2];
  }

  //! Returns the distance squared between this and a.
  float distance2(const mVec3& a) const
  {
    return( mSquare(v[0] - a.v[0]) +
            mSquare(v[1] - a.v[1]) +
            mSquare(v[2] - a.v[2]) );
  }

  //! Returns the distance between this and a.
  float distance(const mVec3& a) const
  {
    return( mSqrt(distance2(a)) );
  }

  //! Returns the length squared of this.
  float length2() const
  {
    return( mSquare(v[0]) + 
            mSquare(v[1]) + 
            mSquare(v[2]) );
  }

  //! Returns the length of this.
  float length() const
  {
    return( mSqrt(length2()) );
  }

  //! Returns the dot product of this and a.
  float dot(const mVec3& a) const
  {
    return( v[0] * a.v[0] +
            v[1] * a.v[1] +
            v[2] * a.v[2] );
  }

  //! Calculates the cross product of a and b
  void cross(const mVec3& a, const mVec3& b)
  {
    mVec3 cp;
    cp.v[0] = a.v[1]*b.v[2] - a.v[2]*b.v[1];
    cp.v[1] = a.v[2]*b.v[0] - a.v[0]*b.v[2];
    cp.v[2] = a.v[0]*b.v[1] - a.v[1]*b.v[0];
    set(cp);
  }

  //! Normalizes this and returns the previous length.
  float normalize()
  {
    float l(length());
    if (l != 0.0f) scale(1.0f / l);
    return l;
  }

  //! Linearly interpolates between two vectors.
  void lerp(const mVec3& a, const mVec3& b, float t)
  {
    v[0] = a.v[0] + t * (b.v[0] - a.v[0]);
    v[1] = a.v[1] + t * (b.v[1] - a.v[1]);
    v[2] = a.v[2] + t * (b.v[2] - a.v[2]);
  }

  //! Creates a normal from three points.
  void makeNormal(const mVec3& a, const mVec3& b, const mVec3& c);

  //! Transforms the point p into f's axial frame.
  void xformTo(const mVec3& p, const mFrame& f);

  //! Transforms the point p into the OCS.
  void xformTo(const mVec3& p, const mOCS& ocs);

  //! Transforms the vector into the OCS.
  /*! Assumes \e vec is a vector.  The ocs's origin is not used. */
  void xformVecTo(const mVec3& vec, const mOCS& ocs);

  //! Transforms the point p from f's axial frame.
  void xformFrom(const mVec3& p, const mFrame& f);

  //! Transforms the point p from the OCS.
  void xformFrom(const mVec3& p, const mOCS& ocs);

  //! Transforms the vector from the OCS.
  /*! Assumes \e vec is a vector.  The ocs's origin is not used. */
  void xformVecFrom(const mVec3& vec, const mOCS& ocs);

};

// Common constant values.
const mVec3 mZero3(0.0f, 0.0f, 0.0f);
const mVec3 mOne3(1.0f, 1.0f, 1.0f);

//****************************************************************************
//**    END HEADER mVec3.h
//****************************************************************************
#endif // __mVec3_H__
